<?php

namespace Drupal\twig_to_sdc\Drush\Commands;

use Drush\Attributes as CLI;
use Drush\Commands\AutowireTrait;
use Drush\Commands\DrushCommands;
use Symfony\Component\Filesystem\Filesystem;

/**
 * Provides Drush commands to generate .component.yml files from Twig templates.
 *
 * This class defines commands to analyze Twig files in a specified directory,
 * parse variable details, and generate corresponding .component.yml files to
 * support Single Directory Components (SDC) in Drupal.
 */
final class SdcGeneratorCommands extends DrushCommands {
  use AutowireTrait;

  /**
   * Drush command to generate .component.yml files from Twig templates.
   *
   * This command scans the specified directory for Twig files, parses variable
   * details using services, and generates corresponding .component.yml files
   * for each Twig file.
   *
   * @param string $path
   *   The directory path to check for .twig files.
   *
   * @command twig_to_sdc:generate
   * @aliases tsdc:gen
   * @usage twig_to_sdc:generate /path/to/directory
   *   Generates .component.yml files for Twig files in the specified directory.
   */
  #[CLI\Command(name: 'twig_to_sdc:generate', aliases: ['tsdc:gen'])]
  #[CLI\Argument(name: 'path', description: 'Directory path to check for .twig files.')]
  #[CLI\Usage(name: 'twig_to_sdc:generate /path/to/directory', description: 'Generate .component.yml files for Twig templates in the directory.')]
  public function commandName($path) {
    $fs = new Filesystem();

    // Check if the provided path exists and is a directory.
    if ($fs->exists($path) && is_dir($path)) {
      $this->logger()->success(dt('The provided directory path is valid: @path', ['@path' => $path]));

      // Count the number of .twig files in the directory.
      $twigFileCount = $this->countTwigFiles($path);

      if ($twigFileCount > 0) {
        $this->logger()->success(dt('Found @count .twig files in the directory.', ['@count' => $twigFileCount]));

        // Iterate through each .twig file and generate .component.yml.
        foreach (new \RecursiveIteratorIterator(new \RecursiveDirectoryIterator($path)) as $file) {
          if ($file->isFile() && $file->getExtension() === 'twig') {
            $componentName = pathinfo($file->getFilename(), PATHINFO_FILENAME);

            try {
              // Read the content of the Twig file.
              $twigContent = file_get_contents($file->getPathname());
              $relativeFilePath = $file->getPathname();
              $this->logger()->success(dt('Analyzing your @file.', ['@file' => $relativeFilePath]));

              // Generate a prompt and parse variables.
              $prompt = \Drupal::service('twig_to_sdc.twig_variable_comment')->generateTwigComment($twigContent);
              $result = \Drupal::service('twig_to_sdc.twig_comment_parser')->parseVariables($prompt, $componentName, $path);

              // Generate the .component.yml file.
              $result = \Drupal::service('twig_to_sdc.sdc_yml_generator')->generateYaml($componentName, $result, $relativeFilePath, $path);
            }
            catch (\Exception $e) {
              $this->logger()->error(dt('Failed to generate examples for your component props.'));
            }
          }
        }
      }
      else {
        $this->logger()->error(dt('No .twig files found in the directory: @path', ['@path' => $path]));
        return;
      }
    }
    else {
      $this->logger()->error(dt('The provided directory path is invalid or could not be accessed: @path', ['@path' => $path]));
    }
  }

  /**
   * Counts the number of .twig files in the specified directory.
   *
   * This method traverses the directory recursively and counts all .twig files.
   *
   * @param string $directory
   *   The directory to search for .twig files.
   *
   * @return int
   *   The total number of .twig files found in the directory.
   */
  protected function countTwigFiles($directory) {
    $iterator = new \RecursiveIteratorIterator(
      new \RecursiveDirectoryIterator($directory)
    );
    $twigFileCount = 0;

    // Iterate through the files and count .twig files.
    foreach ($iterator as $file) {
      if ($file->isFile() && $file->getExtension() === 'twig') {
        $twigFileCount++;
      }
    }

    return $twigFileCount;
  }

}
