<?php

namespace Drupal\twig_to_sdc\services;

use Symfony\Component\Yaml\Yaml;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;

/**
 * Service to generate formatted SDC .component.yml file.
 */
class SdcYamlGenerator {
  /**
   * The logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Constructs the SdcYamlGenerator object.
   *
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory service.
   */
  public function __construct(LoggerChannelFactoryInterface $logger_factory) {
    $this->logger = $logger_factory->get('twig_to_sdc');
  }

  /**
   * Generate YAML data for a Drupal component.
   *
   * @param string $component_name
   *   The name of the component in kebab-case.
   * @param array $result
   *   The processed data including variables, slots, and conditional variables.
   * @param string $filePath
   *   Path to the source Twig file.
   * @param string $path
   *   Directory path for saving the output file.
   */
  public function generateYaml($component_name, $result, $filePath, $path) {
    $variables = $result['variables'];
    $slots = $result['slots'];
    $conditional_variables = $result['conditional_variables'];

    // Determine required fields based on descriptions, default values,
    // and conditionals.
    $required_fields = [];
    foreach ($variables as $key => $var) {
      if (in_array($key, ["attributes"])) {
        continue;
      }
      if (strpos($var["description"], "optional") === FALSE && $var["type"] != "boolean") {
        if (!array_key_exists("default", $var) && !in_array($key, $conditional_variables)) {
          $required_fields[] = $key;
        }
      }
    }

    // Prepare YAML data structure.
    $yaml_data = [
      "name" => ucfirst(str_replace('-', ' ', $component_name)),
      "status" => "experimental",
      "group" => 'default',
      "props" => [
        "type" => "object",
      ],
    ];

    // Add required fields to the 'props' section if they exist.
    if (!empty($required_fields)) {
      $yaml_data["props"]["required"] = $required_fields;
    }

    // Add properties (variables) to 'props'.
    $yaml_data["props"]["properties"] = $variables;

    // Add slots if they exist.
    if (!empty($slots)) {
      $yaml_data["slots"] = $slots;
    }

    // Convert the array to YAML format.
    $yaml_data = Yaml::dump($yaml_data, 20, 2);

    // Get the parent directory.
    $parentDir = dirname($filePath);

    if (is_writable($parentDir)) {
      // Change the extension from .twig to .component.yml.
      $newFilePath = preg_replace('/\.twig$/', '.component.yml', $filePath);

      // Write the content to the new file.
      if (file_put_contents($newFilePath, $yaml_data) !== FALSE) {
        $this->logger->notice(dt("File @newFilePath created successfully", ['@newFilePath' => $newFilePath]));
      }
      else {
        $this->logger->warning(dt("$newFilePath Failed to create the file.", ['@newFilePath' => $newFilePath]));
      }
    }
    else {
      $this->logger->error(dt("The directory '$parentDir' is not writable.", ['@newFilePath' => $parentDir]));
    }
  }

}
