<?php

declare(strict_types=1);

namespace Drupal\twig_to_sdc\services;

use Drupal\ai\AiProviderPluginManager;
use Drupal\ai\OperationType\Chat\ChatInput;
use Drupal\ai\OperationType\Chat\ChatMessage;
use Drupal\Core\Url;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;

/**
 * Class responsible for generating Twig comments and example data using AI.
 */
final class TwigAnnotator {
  /**
   * The AI provider.
   *
   * @var \Drupal\ai\AiProviderInterface|\Drupal\ai\Plugin\ProviderProxy
   */
  protected $aiProvider;

  /**
   * The logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * Constructs a TwigVariableComment object.
   *
   * @param \Drupal\ai\AiProviderPluginManager $aiProvider
   *   The AI provider service.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory service.
   */
  public function __construct(AiProviderPluginManager $aiProvider, LoggerChannelFactoryInterface $logger_factory) {
    $this->aiProvider = $aiProvider;
    $this->logger = $logger_factory->get('twig_to_sdc');
  }

  /**
   * {@inheritdoc}
   */
  public function generateTwigComment($twig_content): string {
    $system_prompt = "You’re a seasoned Twig developer with extensive experience in writing clear and concise comments for Twig templates. Ensure that each variable is explained clearly, including its type and purpose. Focus only on variables with primitive data types as well as arrays and objects, and ignore any other types.
    If a variable has specific options, include them with the label 'options:' on the same line as the description. Keep comments concise yet informative to make it easy for future developers to understand the usage and significance of each variable. Make sure that you exclude variables that are defined locally in twig. For example, the comment on a variable might look like this:
      {#
        /**
         * @file
         * Component name.
         *
         * Variables:
         * - variable: [array<datatype>] Array of link objects containing:
         *   - property1: [string] description.
         *   - property2: [string] description.
         *   - property3: [array<datatype>] description
         *       - property5: [string] description.
         *       - property6: [string] description.
         * - variable2: [string] Content slot.
         * - variable3: [string] Site section.
         * - variable4: [string] The options: radio, checkbox.
         * - variable5: [string] Title.
         * - variable6: [boolean] Show as icon.
         * - image: [object] Featured image object:
         *   - url: [string] Source.
         *   - alt: [string] Alt text.
         * - variable7: [string] Type of the background image blending. Defaults to 'normal'.
         * - variable8: [string] Theme: light, dark.
         */
      #}
      Generate above as the output only.
      ";

    $twig_comment = $this->generateAiResponse($twig_content, $system_prompt);
    $twig_content = $twig_comment . "\n" . $twig_content;
    return $twig_content;
  }

  /**
   * Generates example values for variables that lack examples based on types.
   *
   * @param array $variables
   *   A list of variables for which to generate examples.
   *
   * @return array
   *   The generated examples as an associative array.
   */
  public function generateExamples($variables): array {
    $system_prompt = 'You’re an expert in data modeling and JSON structure, with a specialization in generating realistic examples for various variable types. You have a deep understanding of how to create contextually appropriate values based on variable descriptions and types, ensuring that your outputs are relevant and accurate.

    Your task is to generate an example for a variable that does not already have an example. Return the following JSON object as a single string without formatting or newlines
     
    output response:
    {
      "property1": "example value",
      "property2": "example value"
    }
    ';

    $return_message = $this->generateAiResponse(json_encode($variables), $system_prompt);
    return json_decode($return_message, TRUE);
  }

  /**
   * Generates AI responses based on user input and a system prompt.
   *
   * Sends a request to the AI provider and processes the response.
   *
   * @param string $user_input
   *   The input provided by the user to the AI model.
   * @param string $system_prompt
   *   The instruction to guide the AI's response.
   *
   * @return mixed
   *   The AI's response, typically text or decoded JSON.
   */
  private function generateAiResponse(string $user_input, string $system_prompt): mixed {
    $messages = new ChatInput([new ChatMessage('user', $user_input)]);

    // Get the default provider settings.
    $sets = $this->aiProvider->getDefaultProviderForOperationType('chat');
    if (empty($sets['provider_id'])) {
      $this->logger->error(dt('Please set a default chat provider & model in the <a href=":url">AI module settings</a>.', [':url' => Url::fromRoute('ai.settings_form')->toString()], 'error', TRUE));
    }

    // Create the AI provider instance.
    $provider = $this->aiProvider->createInstance($sets['provider_id']);
    $provider->setChatSystemRole($system_prompt);
    // Send the request and get the response.
    $response = $provider->chat($messages, $sets['model_id'], ['my-custom-call']);
    $return_message = $response->getNormalized();

    // Return the response text or decoded JSON.
    $response_text = $return_message->getText();
    return $response_text;
  }

}
