<?php

namespace Drupal\twig_to_sdc\services;

/**
 * Class TwigCommentParser.
 *
 * Provides functionality to parse variables, slots, conditionals, and includes
 * from Twig templates for Single-Directory Components (SDC).
 */
class TwigCommentParser {

  /**
   * Parses variables, slots, conditionals, and includes from Twig content.
   *
   * @param string $twig_content
   *   The content of the Twig file.
   * @param string $component_name
   *   The name of the component.
   * @param string $file_directory
   *   Directory of the Twig file.
   *
   * @return array
   *   Parsed data including variables, slots, and conditional variables.
   */
  public function parseVariables($twig_content, $component_name, $file_directory) {
    // Initialize arrays for parsed data.
    $slots = [];
    // Extract variables and their metadata from Twig content.
    $variables = $this->extractVariables($twig_content);
    // Check for variables used in include statements.
    $include_slots = $this->checkVariableInIncludes($twig_content, $variables);
    // Merge slots extracted from include statements.
    $slots = array_merge($slots, $this->extractSlots($variables, $include_slots));
    // Assign default values to variables, if available.
    $this->assignDefaults($twig_content, $variables);
    // Extract variables used in conditional logic.
    $conditional_variables = $this->extractConditionalVariables($twig_content);
    // Add generated examples to variables.
    $this->mergeExamples($variables);

    // Return the parsed data.
    return [
      'variables' => $variables,
      'slots' => $slots,
      'conditional_variables' => $conditional_variables,
    ];
  }

  /**
   * Extracts variables defined in Twig comments.
   *
   * @param string $twig_content
   *   The content of the Twig file.
   * @param int $nested
   *   Current nesting level for parsing variables.
   *
   * @return array
   *   Extracted variables with their metadata.
   */
  private function extractVariables($twig_content, $nested = 1) {
    $variables = [];
    $variable_pattern = "/\* {" . $nested . "}- (\w+): \[(\w+|object|array<object>)\] (.*)/";

    // Match variable declarations in Twig comments.
    if (preg_match_all($variable_pattern, $twig_content, $matches, PREG_SET_ORDER)) {
      foreach ($matches as $match) {
        $entry = $this->buildVariableEntry($match, $twig_content, $nested);
        if (!empty($entry)) {
          $variables[$match[1]] = $entry;
        }
      }
    }

    return $variables;
  }

  /**
   * Build a variable entry with type, title, and other metadata.
   *
   * This function processes a variable's details, extracting its type,
   * description, and additional metadata such as enum values, examples,
   * and handling complex types like objects and arrays.
   *
   * @param array $match
   *   The regex match array containing variable details.
   * @param string $twig_content
   *   The content of the Twig file where the variable is defined.
   * @param bool $nested
   *   Whether the variable is nested (inside an object or array).
   *
   * @return array
   *   An associative array representing the variable metadata, or an empty
   *   array if the variable should not be processed.
   */
  private function buildVariableEntry($match, $twig_content, $nested) {
    // Extract the variable name, type, and description from the match array.
    $var_name = $match[1];
    $var_type = $match[2];
    $var_desc = str_replace(["'", '"'], "", $match[3]);

    // Define a pattern to extract enum values from variable comments.
    $enum_pattern = "/\* - (\w+): \[(string)\] .*?: ([^.,]+(?:, [^.,]+)*)/";

    // Skip processing for the special "attributes" variable.
    if ($var_name == 'attributes') {
      return [];
    }

    // Initialize the variable entry with basic metadata.
    $entry = [
      'type' => $var_type !== 'object' ? $var_type : 'object',
      'title' => ucwords(str_replace('_', ' ', $var_name)),
      'description' => $var_desc,
    ];

    // Check for and handle enum values defined in the comments.
    if (preg_match($enum_pattern, $match[0], $enum_match)) {
      $entry['enum'] = array_map('trim', explode(',', $enum_match[3]));
      $entry['examples'] = $entry['enum'];
    }

    // Handle specific types with custom logic.
    if ($var_type === 'boolean') {
      // Boolean variables have predefined examples and no description.
      $entry['examples'] = [TRUE, FALSE];
      unset($entry['description']);
    }
    elseif ($var_type === 'object') {
      // Handle nested object variables by delegating to a helper function.
      $this->handleObjectVariables($twig_content, $var_name, $entry, $nested);
    }
    elseif (str_starts_with($var_type, 'array')) {
      // Process array types to determine their inner type and metadata.
      $entry['type'] = 'array';
      $array_type = '';
      $pattern = '/array<([^>]+)>/';

      // Extract the array's inner type if specified.
      if (preg_match($pattern, $var_type, $matches)) {
        $array_type = $matches[1];
      }
      // Delegate array variable processing to a helper function.
      $this->handleArrayVariables($twig_content, $var_name, $array_type, $entry, $nested);
    }

    // Return the constructed variable metadata entry.
    return $entry;
  }

  /**
   * Process object variables, checking includes or parsing properties.
   *
   * This function processes object variables by examining their scope in
   * the provided Twig content. If the object has properties, they are
   * parsed and added to the variable entry.
   *
   * @param string $twig_content
   *   The content of the Twig file to search for object variable details.
   * @param string $var_name
   *   The name of the object variable being processed.
   * @param array &$entry
   *   Reference to the entry array where the processed data is added.
   * @param int $nested
   *   The current nesting level for the variable being processed.
   */
  private function handleObjectVariables($twig_content, $var_name, &$entry, $nested) {
    // Define the regex pattern to match the scope of the object variable.
    $objectScopePattern = "/\* {" . $nested . "}- " . preg_quote($var_name, '/') . ": \\[object\\](.*?)(\\* - |\\Z)/s";

    // If the pattern matches, parse the object properties from the content.
    if (preg_match($objectScopePattern, $twig_content, $objectScopeMatch)) {
      $entry['properties'] = $this->parseObjectProperties($objectScopeMatch[1], $var_name, $nested);
    }
  }

  /**
   * Process array variables by extracting items and their properties.
   *
   * This function handles array variables by identifying their scope and
   * extracting the properties of the array items. If the array type is an
   * object, it processes those properties recursively.
   *
   * @param string $twig_content
   *   The content of the Twig file to search for array variable details.
   * @param string $var_name
   *   The name of the array variable being processed.
   * @param string $array_type
   *   The type of items contained in the array (e.g., 'object').
   * @param array &$entry
   *   Reference to the entry array where the processed data is added.
   * @param int $nested
   *   The current nesting level for the variable being processed.
   */
  private function handleArrayVariables($twig_content, $var_name, $array_type, &$entry, $nested) {
    // Define the regex pattern to match the scope of the array variable.
    $arrayScopePattern = "/\* {" . $nested . "}- " . preg_quote($var_name, '/') . ": \\[array<" . $array_type . ">\\](.*?)(\\* - |\\Z)/s";

    // If the pattern matches, parse the array properties from the content.
    if (preg_match($arrayScopePattern, $twig_content, $arrayScopeMatch)) {
      $entry['items'] = $this->parseArrayProperties($arrayScopeMatch[1], $array_type, $nested);
    }
  }

  /**
   * Extract properties from an array scope.
   *
   * This function extracts the properties of array items from the array scope
   * content. It handles both object and non-object array types, adjusting the
   * format of the properties accordingly.
   *
   * @param string $arrayScopeContent
   *   The content within the array scope to be parsed.
   * @param string $array_type
   *   The type of array items (e.g., 'object').
   * @param int $nested
   *   The current nesting level for the array items being processed.
   *
   * @return array|null
   *   An associative array with the properties of the array items, or NULL
   *   if no properties are found.
   */
  private function parseArrayProperties($arrayScopeContent, $array_type, $nested) {
    // Increase the nesting level for the properties inside the array.
    $nested = $nested + 2;

    // Extract the variables (properties) from the array scope content.
    $properties = $this->extractVariables($arrayScopeContent, $nested);

    // If the type is 'object' and has properties, return as object type.
    if ($array_type == 'object' && !empty($properties)) {
      return ['type' => 'object', 'prefixItems' => $properties];
    }
    // For non-object arrays, return array type or NULL if no properties found.
    else {
      return !empty($properties) ? ['type' => $array_type] : NULL;
    }
  }

  /**
   * Parse object properties from the Twig content.
   *
   * This function parses the object properties for a given object variable
   * from the provided Twig content. It uses a regular expression to match
   * the properties defined within the object scope and extracts the
   * relevant details for further processing.
   *
   * @param string $twig_content
   *   The content of the Twig file, which includes the object variable details.
   * @param string $var_name
   *   The name of the object variable being processed.
   * @param int $nested
   *   The current nesting level of the variable in the Twig content.
   *
   * @return array
   *   An array of parsed properties for the object variable.
   */
  protected function parseObjectProperties($twig_content, $var_name, $nested) {
    // Increment the nesting level to identify properties inside the object.
    $nested = $nested + 2;

    // Define the pattern to match the object variable and its properties.
    $pattern = "/\* {" . $nested . "}- (\w+): \[(\w+)\] (.*)/";

    // Extract the properties of the object using the extractVariables method.
    $properties = $this->extractVariables($twig_content, $nested);

    // Return the parsed properties.
    return $properties;
  }

  /**
   * Assign default values to variables if specified in Twig content.
   *
   * This function checks for Twig-specific default assignments in the content,
   * such as variables with default values provided through the `set` tag
   * or the `default` filter. If any defaults are found, they are added
   * to the corresponding variable entries.
   *
   * @param string $twig_content
   *   The content of the Twig file, which may contain default assignments.
   * @param array &$variables
   *   A reference to the variables array, where default values will be added
   *   to the corresponding variables.
   */
  private function assignDefaults($twig_content, &$variables) {
    // Define patterns to match the default assignments in the Twig content.
    $patterns = [
    // Matches set assignments with default values.
      "/{% set (\w+) = [^%]+ \? [^:]+ : (.+?) %}/",
    // Matches the default filter.
      "/(\w+)\|default\('(.+?)'\)/",
    ];

    // Loop through the defined patterns to find matches in the Twig content.
    foreach ($patterns as $pattern) {
      // If matches are found, process each match.
      if (preg_match_all($pattern, $twig_content, $matches, PREG_SET_ORDER)) {
        foreach ($matches as $match) {
          // Extract the variable name and default value.
          $var_name = $match[1];
          $default_value = trim($match[2], "'\"");

          // If the variable exists, assign its default value.
          if (isset($variables[$var_name])) {
            $variables[$var_name]['default'] = $this->parseDefaultValue($default_value);
          }
        }
      }
    }
  }

  /**
   * Parse and return the correct data type for default values.
   *
   * This function processes default values in the Twig content.
   * It converts common string representations of boolean and null values to
   * their corresponding PHP types (e.g., 'true' to TRUE, 'false' to FALSE).
   *
   * @param string $default_value
   *   The default value as a string to be parsed into the correct data type.
   *
   * @return mixed
   *   The parsed default value: boolean, null, or original value.
   */
  private function parseDefaultValue($default_value) {
    // Return NULL if the default value is 'null'.
    if ($default_value === 'null') {
      return NULL;
    }

    // Return FALSE if the default value is 'false'.
    if ($default_value === 'false') {
      return FALSE;
    }

    // Return TRUE if the default value is 'true'.
    if ($default_value === 'true') {
      return TRUE;
    }

    // Return the default value as-is if it's not 'null', 'false', or 'true'.
    return $default_value;
  }

  /**
   * Extract slot variables from the variables list.
   *
   * This function identifies and extracts "slot" variables from the list of
   * variables. Slot variables are marked by the presence of the word 'slot'
   * in their description. It also processes variables included in specific
   * slots defined by include statements in the Twig content.
   *
   * @param array &$variables
   *   A reference to the variables array, from which slot variables will be
   *   extracted.
   * @param array $include_slots
   *   An associative array of slot names and their associated variables
   *   defined in include statements.
   *
   * @return array
   *   An array of extracted slot variables with their titles and descriptions.
   */
  private function extractSlots(&$variables, $include_slots) {
    // Initialize an empty array to hold the extracted slot variables.
    $slots = [];

    // Extract slot variables from the array based on their descriptions.
    foreach ($variables as $key => $value) {
      if (isset($value['description']) && stripos($value['description'], 'slot') !== FALSE) {
        // Add the slot variable to the $slots array.
        $slots[$key] = $value;
        // Remove the slot variable from the main variables array.
        unset($variables[$key]);
      }
    }

    // Process variables defined in include statements.
    foreach ($include_slots as $slot_name => $slot_variables) {
      // Format the slot name and title.
      $formatted_slot_name = str_replace('-', '_', $slot_name);
      $formatted_title = ucfirst(str_replace('-', ' ', $slot_name)) . ' slot.';
      $description = 'Slot created for include statement for file ' . $slot_name . '.twig.';

      foreach ($slot_variables as $variable) {
        // Check if the variable exists in the variables array.
        if (isset($variables[$variable])) {
          // Add the slot variable with title and description to $slots array.
          $slots[$formatted_slot_name] = [
            'title' => $formatted_title,
            'description' => $description,
          ];
          // Remove the variable from the main variables array after processing.
          unset($variables[$variable]);
        }
      }
    }

    // Return the array of extracted slot variables.
    return $slots;
  }

  /**
   * Extract conditional variables used in Twig logic.
   *
   * This function identifies variables that are involved in conditional logic
   * within the Twig content, such as variables used in {% if %} statements
   * or variables that are checked for null using the "??" operator.
   *
   * @param string $twig_content
   *   The content of the Twig file, which may contain conditional logic.
   *
   * @return array
   *   An array of unique variables used in conditional expressions.
   */
  private function extractConditionalVariables($twig_content) {
    // Initialize an array to store the conditional variables.
    $conditional_vars = [];

    // Define the patterns to match {% if %} and null-coalescing operations.
    $patterns = [
    // Matches variables in {% if %} statements.
      "/{% if (\w+) %}/",
    // Matches variables checked for null with the "??" operator.
      "/\b(\w+)\s*\?\?\s*null\b/",
    ];

    // Loop through the patterns and find matches in the Twig content.
    foreach ($patterns as $pattern) {
      if (preg_match_all($pattern, $twig_content, $matches)) {
        // Merge the found variables into the $conditional_vars array.
        $conditional_vars = array_merge($conditional_vars, $matches[1]);
      }
    }

    // Return only the unique conditional variables.
    return array_unique($conditional_vars);
  }

  /**
   * Merge examples into variables, prioritizing generated examples.
   *
   * This function generates example values for the variables if they are not
   * already defined. The examples are added to the variables, but only if
   * the variable type is not "array" or "object". The generated examples
   * take precedence over any existing examples.
   *
   * @param array &$variables
   *   A reference to the variables array, where the examples will be merged.
   */
  private function mergeExamples(&$variables) {
    // Generate variable examples using the twig_variable_comment service.
    $examples = \Drupal::service('twig_to_sdc.twig_variable_comment')->generateExamples($variables);

    // Loop through examples and merge them into the variables array.
    foreach ($examples as $key => $value) {
      // Check if the variable exists and doesn't already have examples.
      if (isset($variables[$key]) && !isset($variables[$key]['examples'])) {
        // Skip variables of type "array" or "object".
        // @todo find out examples and update this.
        if (str_starts_with($variables[$key]['type'], 'array') || in_array($variables[$key]['type'], ['object'])) {
          continue;
        }

        // Set examples for the variable, using the default value if present.
        $variables[$key]['examples'] = [$variables[$key]['default'] ?? $value];
      }
    }
  }

  /**
   * Check if a variable or its properties are in include values.
   *
   * @param string $twig_content
   *   The content of the Twig file.
   * @param array $variables
   *   The list of variables to search for in the include statements.
   *
   * @return array
   *   Tuple with the Twig file name and found variables in the include block.
   *   Returns an empty array if not found.
   */
  protected function checkVariableInIncludes($twig_content, $variables) {
    // Pattern to match variable names and their properties.
    $variable_pattern = '/["\']?(\w+)["\']?\s*:\s*([\w]+)(?:\.[\w\[\]]+)?|\'[^\']*\'|"[^"]*"/';

    $include_result = [];

    // Reduce the list of variables to include only those with a defined type.
    $variables = array_reduce(array_keys($variables), function ($result, $key) use ($variables) {
      $item = $variables[$key];
      // Ensure both indices exist.
      if (isset($item['type'])) {
          $result[$key] = $item['type'];
      }
      return $result;
    }, []);

    // Check for direct include statements in the content.
    if ($result = $this->processIncludeStatements($twig_content, $variables, $variable_pattern)) {
      $include_result[] = $result;
    }

    // Check for indirect variable usage in loop blocks.
    if ($result = $this->processLoopStatements($twig_content, $variables, $variable_pattern)) {
      $include_result[] = $result;
    }

    // Merge all results from both direct and indirect checks.
    $include_result = array_merge(...$include_result);
    $slot_variables = [];

    // Iterate through include results to identify and process slot variables.
    foreach ($include_result as $key => $result) {
      $slot = FALSE;
      // Skip if no variables are found in the result.
      if (!isset($result['variables'])) {
        continue;
      }

      // Check if the result contains array or object variables (slots).
      foreach ($result['variables'] as $variable_type) {
        if (in_array($variable_type, ['array', 'object'])) {
          $slot = TRUE;
          break;
        }
      }

      // If a slot variable is found, categorize and store the slot variables.
      if ($slot) {
        if (count($result['variables']) == 1) {
          $slot_id = array_keys($result['variables'])[0];
          $slot_variables[$slot_id] = array_keys($result['variables']);
        }
        else {
          // Create a unique identifier for slot variables.
          if (isset($slot_variables[$result['component']])) {
            $slot_variables[$result['component'] . '_' . $key] = array_keys($result['variables']);
          }
          else {
            $slot_variables[$result['component']] = array_keys($result['variables']);
          }
        }
      }
    }

    return $slot_variables;
  }

  /**
   * Process loops and check for includes within the loop block.
   *
   * @param string $twig_content
   *   The content of the Twig file.
   * @param array $variables
   *   The list of variables to match against.
   * @param string $variable_pattern
   *   The regex pattern for variables in 'with' clause.
   *
   * @return array|null
   *   An array of include statement results if found, NULL otherwise.
   */
  private function processLoopStatements($twig_content, $variables, $variable_pattern) {
    // Regex pattern to match loop statements.
    $loop_pattern = '/{%\s*for\s+(\w+)\s+in\s+(\w+)\s*%}/';

    // Check for loop matches in the Twig content.
    if (preg_match_all($loop_pattern, $twig_content, $loop_matches, PREG_SET_ORDER)) {
      foreach ($loop_matches as $loop_match) {
        // Extract loop variables.
        $loop_var = $loop_match[1];
        $loop_source = $loop_match[2];

        // If the loop source exists in the variables, process its content.
        if (in_array($loop_source, array_keys($variables))) {
          // Extract content inside the loop block.
          $loop_content_start = strpos($twig_content, $loop_match[0]) + strlen($loop_match[0]);
          $loop_content_end = strpos($twig_content, '{% endfor %}', $loop_content_start);
          $loop_content = substr($twig_content, $loop_content_start, $loop_content_end - $loop_content_start);

          // Check for include statements within the loop block.
          if ($result = $this->processIncludeStatements($loop_content, $variables, $variable_pattern, [$loop_var => $loop_source])) {
            return $result;
          }
        }
      }
    }
    return NULL;
  }

  /**
   * Process include statements and search for the target variable names used.
   *
   * @param string $twig_content
   *   The Twig template content.
   * @param array $variables
   *   The list of variables to match against.
   * @param string $variable_pattern
   *   Regex pattern for variables in the 'with' clause of includes.
   * @param array $loop_source
   *   The loop source variables to consider while matching.
   *
   * @return array
   *   A structured array with include statement details and matched variables.
   */
  private function processIncludeStatements($twig_content, $variables, $variable_pattern, $loop_source = []) {
    $include_data = [];

    // Regex pattern to match include statements with the 'with' clause.
    $include_statement_pattern = '/\{%\s*include\s*["\']([^"\']+)["\']\s*with\s*\{\s*([^}]*?)\s*\}\s*(only\s*)?%\}/';

    // Regex to match the include statement syntax.
    $include_pattern = '/{%\s*include\s*["\']@?(?:[^\/]+\/)*([^"\']+\.twig)["\']\s*with\s*\{\s*([\s\S]*?)\}\s*(only)?\s*%}/';

    // Match all include statements.
    preg_match_all($include_statement_pattern, $twig_content, $include_statement_matches);

    foreach ($include_statement_matches[0] as $include_statement_match) {
      // Match variables in the 'with' clause of include statements..
      if (preg_match_all($include_pattern, $include_statement_match, $matches, PREG_SET_ORDER)) {
        foreach ($matches as $include_match) {
          // Extract the Twig file name and variables inside the 'with' clause.
          $file_name = $include_match[1];
          $variables_content = $include_match[2];

          // Extract variables from the 'with' clause.
          $variables_dict = $this->extractIncludeVariables($variables_content, $variable_pattern);

          // Match extracted variables against the provided list.
          $matched_variables = [];

          if (count($loop_source)) {
            foreach ($variables_dict as $var_name => $var_value) {
              if (in_array($var_value, array_keys($loop_source))) {
                $matched_variables[$loop_source[$var_value]] = $variables[$loop_source[$var_value]];
              }
            }
          }
          else {
            foreach ($variables_dict as $var_name => $var_value) {
              if (in_array($var_value, array_keys($variables), TRUE)) {
                $matched_variables[$var_value] = $variables[$var_value];
              }
            }
          }

          // Store matched include data if variables are found.
          if (!count($matched_variables)) {
            continue;
          }

          $include_data[] = [
            'component' => str_replace('.twig', '', $file_name),
            'variables' => $matched_variables,
          ];
        }
      }
    }

    // Return the structured include data.
    return $include_data;
  }

  /**
   * Extract variables from the 'with' content of an include statement.
   *
   * @param string $variables_content
   *   The content inside the 'with' clause of an include statement.
   * @param string $variable_pattern
   *   The regex pattern to match variables.
   *
   * @return array
   *   A dictionary of variable names and their corresponding values.
   */
  private function extractIncludeVariables($variables_content, $variable_pattern) {
    $variables_dict = [];
    // Use regex to extract variables from the 'with' content.
    if (preg_match_all($variable_pattern, $variables_content, $var_matches, PREG_SET_ORDER)) {
      foreach ($var_matches as $var_match) {
        if (isset($var_match[1]) && isset($var_match[2])) {
          $var_name = trim($var_match[1]);
          $var_value = trim($var_match[2]);
          $variables_dict[$var_name] = $var_value;
        }
      }
    }
    return $variables_dict;
  }

}
